//========================================================================================
//
//	File:				Selection.cpp
//	Release Version:	$ ODF 2 $ 
//
//	Copyright:			(c) 1993 - 1996 by Apple Computer, Inc., all rights reserved.
//
//========================================================================================

#include "Table.hpp"

// ----- TablePart Includes -----

#ifndef SELECTION_H
#include "Selection.h"
#endif

#ifndef PROXY_H
#include "Proxy.h"
#endif

#ifndef FRAME_H
#include "Frame.h"
#endif

#ifndef LINKING_H
#include "Linking.h"
#endif

#ifndef CONTENT_H
#include "Content.h"
#endif

#ifndef VIEW_H
#include "View.h"
#endif

// ----- ODF Includes -----

#ifndef FWFRAME_H
#include "FWFrame.h"
#endif

#ifndef FWRECT_H
#include "FWRect.h"
#endif

#ifndef FWUTIL_H
#include "FWUtil.h"
#endif

#ifndef FWPRESEN_H
#include "FWPresen.h"
#endif

#ifndef FWITERS_H
#include "FWIters.h"
#endif

#ifndef FWCONTXT_H
#include "FWContxt.h"
#endif

// ----- OS Layer -----

#ifndef FWODGEOM_H
#include "FWODGeom.h"
#endif

#ifndef FWEVENT_H
#include "FWEvent.h"
#endif

// ----- OpenDoc Includes -----

#ifndef SOM_ODShape_xh
#include <Shape.xh>
#endif

#ifndef SOM_ODTransform_xh
#include <Trnsform.xh>
#endif

#ifndef SOM_ODSession_xh
#include <ODSessn.xh>
#endif

//========================================================================================
// RunTime Info
//========================================================================================

#ifdef FW_BUILD_MAC
#pragma segment odfTable
#endif

FW_DEFINE_AUTO(CTableSelection)

//========================================================================================
//	class CTableSelection
//========================================================================================

//----------------------------------------------------------------------------------------
//	CTableSelection::CTableSelection
//----------------------------------------------------------------------------------------

CTableSelection::CTableSelection(Environment* ev, CTablePart* tablePart, CTablePartContent* content) :
	FW_CSelection(ev, true, true),	// allow linking
	fTablePart(tablePart),
	fTableContent(content)
{
	fSelectionContent = FW_NEW(CTableSelectionContent, (ev, tablePart, content));
}

//----------------------------------------------------------------------------------------
//	CTableSelection::~CTableSelection
//----------------------------------------------------------------------------------------

CTableSelection::~CTableSelection()
{
}

//----------------------------------------------------------------------------------------
//	CTableSelection::GetSelectedContent
//----------------------------------------------------------------------------------------

FW_CContent* CTableSelection::GetSelectedContent(Environment* ev)
{
FW_UNUSED(ev);
	return fSelectionContent;
}

//----------------------------------------------------------------------------------------
//	CTableSelection::IsEmpty
//----------------------------------------------------------------------------------------

FW_Boolean CTableSelection::IsEmpty(Environment* ev) const
{
	return fSelectionContent->GetSelectedProxy(ev) == NULL;
}

//----------------------------------------------------------------------------------------
//	CTableSelection::CloseSelection
//----------------------------------------------------------------------------------------

void CTableSelection::CloseSelection(Environment* ev)
{
FW_UNUSED(ev);
}

//----------------------------------------------------------------------------------------
//	CTableSelection::ClearSelection
//----------------------------------------------------------------------------------------

void CTableSelection::ClearSelection(Environment* ev)
{
	CTableProxy* proxy = fSelectionContent->GetSelectedProxy(ev);
	if (proxy)
	{
		//--- Remove the proxy ---
		proxy->DetachEmbeddedFrames(ev);
		fTableContent->RemoveProxy(proxy);
		InvalidateSelection(ev);
	}
}

//----------------------------------------------------------------------------------------
//	CTableSelection::SelectAll
//----------------------------------------------------------------------------------------

void CTableSelection::SelectAll(Environment* ev)
{
FW_UNUSED(ev);
}

//----------------------------------------------------------------------------------------
//	CTableSelection::Select
//----------------------------------------------------------------------------------------

void CTableSelection::Select(Environment* ev, const CCell& cell)
{
	// Get the active frame
	CTableFrame* activeFrame = (CTableFrame*)fTablePart->GetLastActiveFrame(ev);

	// Get the previous selected cell and proxy
	CCell oldCell = fSelectionContent->GetCell();
	CTableProxy* proxy = fTableContent->CellToProxy(oldCell);

	// Get the link manager so we can check for links
	CTableLinkManager* linkMgr = (CTableLinkManager*)fTablePart->GetLinkManager(ev);
	CTableLink* link = NULL;
	CTableLinkSource* linkSource = NULL;

	// if the previous selected cell was containing a proxy then change its selected state
	// and change its hilite
	if (proxy)
	{
		proxy->SetSelectState(ev, FALSE);

		link = linkMgr->CellToLink(ev, oldCell);
		if (link)
			link->Select(ev, false);
		linkSource = linkMgr->CellToSourceLink(ev, oldCell);
		if (linkSource)
			linkSource->Select(ev, false);
	}
	
	// Erase the cell hilite
	if (activeFrame)
		activeFrame->GetTableView(ev)->ChangeHighlightState(ev, kODNoHighlight);

	// select the new cell
	fSelectionContent->SetCell(cell);
	
	// get the proxy contained in the new selected cell
	proxy = fTableContent->CellToProxy(cell);
	
	// if the new selected cell contains a proxy then change its selected state
	// and hilite
	if (proxy)
	{
		proxy->SetSelectState(ev, TRUE);

		link = linkMgr->CellToLink(ev, cell);
		if (link)
			link->Select(ev, true);
		linkSource = linkMgr->CellToSourceLink(ev, cell);
		if (linkSource)
			linkSource->Select(ev, true);
	}

	// Draw the cell hilite
	if (activeFrame)
		activeFrame->GetTableView(ev)->ChangeHighlightState(ev, activeFrame->HasSelectionFocus(ev) ? kODFullHighlight : kODDimHighlight);	
}

//----------------------------------------------------------------------------------------
//	CTableSelection::AcquireSelectionShape
//----------------------------------------------------------------------------------------

ODShape* CTableSelection::AcquireSelectionShape(Environment* ev, ODFacet* facet, FW_CFrame* frame)
{
	FW_UNUSED(facet);
	FW_UNUSED(frame);

	return fSelectionContent->AcquireSuggestedFrameShape(ev);
}

//----------------------------------------------------------------------------------------
//	CTableSelection::InvalidateSelection
//----------------------------------------------------------------------------------------

void CTableSelection::InvalidateSelection(Environment* ev)
{
	FW_CAcquiredODShape aqShape = fSelectionContent->AcquireSuggestedFrameShape(ev);
	GetPresentation(ev)->Invalidate(ev, aqShape);
}

//----------------------------------------------------------------------------------------
// CTableSelection::SelectProxy
//----------------------------------------------------------------------------------------

void CTableSelection::SelectProxy(Environment* ev, CTableProxy* proxy)
{
	CCell cell = proxy->GetCell();
	this->Select(ev, cell);
}

//----------------------------------------------------------------------------------------
// CTableSelection::GetSelectedCell
//----------------------------------------------------------------------------------------

CCell CTableSelection::GetSelectedCell() const
{
	return fSelectionContent->GetCell();
}

//----------------------------------------------------------------------------------------
//	CTableSelection::IsMouseInDraggableItem
//----------------------------------------------------------------------------------------

FW_Boolean CTableSelection::IsMouseInDraggableItem(Environment* ev, 
												FW_CFrame* frame, 
												const FW_CMouseEvent& theMouseEvent, 
												FW_Boolean inBackground)
{
	FW_UNUSED(inBackground);

	CCell cell;
	ETableLoc tl = fTableContent->HitTest(ev, theMouseEvent, frame->GetContentView(ev), cell);
	
	return (tl == kTLCell && fTableContent->CellToProxy(cell) != NULL);
}

//----------------------------------------------------------------------------------------
// CTableSelection::UpdateSelectionOnMouseDown
//----------------------------------------------------------------------------------------

void CTableSelection::UpdateSelectionOnMouseDown(Environment* ev, 
											const FW_CMouseEvent& mouseEvent,
											ODFacet* embeddedFacet,
											FW_Boolean inEmbeddedFrameBorder,
											FW_Boolean inBackground)
{
	FW_UNUSED(inBackground);

	CTableFrame* tableFrame = (CTableFrame*)FW_CFrame::ODtoFWFrame(ev, mouseEvent.GetFacet(ev)->GetFrame(ev));
	
	if (tableFrame->GetTableView(ev)->IsGridShown(ev))
	{
		FW_Boolean select = FALSE;
		CCell cell;
		
		if (inEmbeddedFrameBorder)
		{
			FW_MProxy* proxy = tableFrame->GetProxy(ev, embeddedFacet->GetFrame(ev));
			FW_ASSERT(proxy);
			
			cell = ((CTableProxy*)proxy)->GetCell();	
			select = TRUE;
		}
		else
		{
			ETableLoc tl = fTableContent->HitTest(ev, mouseEvent, tableFrame->GetContentView(ev), cell);	
			select = (tl == kTLCell && cell != GetSelectedCell());
		}
		
		if (select)
			Select(ev, cell);
	}
}

//----------------------------------------------------------------------------------------
//	CTableSelection::CanPasteAsLink
//---------------------------------------------------------------------------------------

FW_Boolean CTableSelection::CanPasteAsLink(Environment* ev, ODPasteAsMergeSetting& setting,
											ODStorageUnit* su)
{
#ifndef FW_DEBUG
FW_UNUSED(ev);
#endif
FW_UNUSED(su);
	setting = kODPasteAsEmbedOnly;	// table has no content to merge

	// we shouldn't be here at all unless the selection is empty
	
	FW_ASSERT(this->IsEmpty(ev));
	return TRUE;
}

//----------------------------------------------------------------------------------------
//	CTableSelection::IsSelectionLinkable
//----------------------------------------------------------------------------------------

FW_Boolean CTableSelection::IsSelectionLinkable(Environment* ev)
{
	FW_Boolean result = fAllowLinkSource;

	if (fAllowLinkSource)
	{
		// make sure the cell is not a link destination
		CTableLinkManager* linkMgr = (CTableLinkManager*)fTablePart->GetLinkManager(ev);
		if (linkMgr->CellToLink(ev, this->GetSelectedCell()))
			result = FALSE;
	}

	return result;
}

//----------------------------------------------------------------------------------------
//	CTableSelection::DoFindLinkSource
//----------------------------------------------------------------------------------------

FW_CLinkSource* CTableSelection::DoFindLinkSource(Environment* ev)	// Override
{
	CTableLinkManager* linkMgr = (CTableLinkManager*)fTablePart->GetLinkManager(ev);
	return linkMgr->CellToSourceLink(ev, fSelectionContent->GetCell());
}

//----------------------------------------------------------------------------------------
//	CTableSelection::GetLinkDestination
//----------------------------------------------------------------------------------------

FW_CLinkDestination* CTableSelection::GetLinkDestination(Environment* ev, FW_Boolean& multipleLinks)
{
FW_UNUSED(multipleLinks);
	// check whether the selected part is in a link destination
	CTableLinkManager* linkMgr = (CTableLinkManager*)fTablePart->GetLinkManager(ev);
	return linkMgr->CellToLink(ev, this->GetSelectedCell());
}

//----------------------------------------------------------------------------------------
//	CTableSelection::GetSavedLinkSpecID
//----------------------------------------------------------------------------------------
ODUpdateID CTableSelection::GetSavedLinkSpecID(Environment* ev)
{
	// Return the link spec ID saved in the currently selected proxy
	CTableProxy* selectedProxy = fSelectionContent->GetSelectedProxy(ev);
	return selectedProxy->GetLinkSpecID();
}

//----------------------------------------------------------------------------------------
//	CTableSelection::SaveLinkSpecID
//----------------------------------------------------------------------------------------
void CTableSelection::SaveLinkSpecID(Environment* ev, ODUpdateID linkSpecID)
{
	// Save the link spec ID in the currently selected proxy
	CTableProxy* selectedProxy = fSelectionContent->GetSelectedProxy(ev);
	selectedProxy->SetLinkSpecID(linkSpecID);
}
